﻿using Microsoft.Office.Interop.Outlook;
using System.Runtime.InteropServices;
using Outlook = Microsoft.Office.Interop.Outlook;
using Word = Microsoft.Office.Interop.Word;

namespace MyOutlookAddin1
{
    class ContextInfo
    {
        public string WindowType { get; private set; } 
        public string WindowCaption { get; private set; }
        public string ContextDataString { get; private set; }
        public string FolderPath { get; private set; }

        public ContextInfo(object context)
        {
            if (context == null) throw new System.ArgumentNullException("context");
            GetSummaryFromContexObject(context);
        }
        public ContextInfo()
        {
            object window = null;
            try
            {
                window = MyOutlookAddin1.AddinModule.CurrentInstance.OutlookApp.ActiveWindow();
                GetSummaryFromContexObject(window);
            }
            catch (System.Exception ex)
            {
                Logging.Log.Add("ContextInfo.ctor", ex);
            }
            finally
            {
                if (window != null) Marshal.ReleaseComObject(window);
            }
        }
        private void GetSummaryFromContexObject(object context)
        {
            if (context is Outlook.Inspector)
            {
                WindowType = "Inspector";
                WindowCaption = (context as Outlook._Inspector).Caption;
                ContextDataString = GetSummaryFromInspector(context as Outlook._Inspector);
            }
            else if (context is Outlook.Explorer)
            {
                WindowType = "Explorer";
                WindowCaption = (context as Outlook._Explorer).Caption;
                ContextDataString = GetSummaryFromExplorer(context as Outlook._Explorer);
            }
            else if (context is Outlook.Selection)
            {
                WindowType = "Explorer";
                Outlook._Explorer explorer = (context as Outlook.Selection).Parent as Outlook._Explorer;
                WindowCaption = explorer.Caption;
                FolderPath = GetFolderPath(explorer);
                ContextDataString = GetSummaryFromSelection(context as Outlook.Selection, explorer);
            }
            else
            {
                // Microsoft.VisualBasic.dll reference is required
                string contextTypeName = Microsoft.VisualBasic.Information.TypeName(context);
                string message = string.Format("{0} - non-supported context type", contextTypeName);
                Logging.Log.Add("ContextInfo.GetSummaryFromContexObject", message);
            }
        }
        private string GetSummaryFromInspector(Outlook._Inspector inspector)
        {
            string summary = null;
            object item = null;
            Word.Document document = null;
            try
            {
                item = inspector.CurrentItem;
                summary = GetSummaryFromOutlookItem(item);
                document = inspector.WordEditor as Word.Document;
                string selectedText = GetTextSelectedInDocument(document);
                if (selectedText == null)
                {
                    summary += "\nSelectedText: (none)";
                }
                else
                {
                    summary += "\nSelectedText='" + selectedText + "'";
                }
            }
            catch (System.Exception ex)
            {
                Logging.Log.Add("ContextInfo.GetData(Inspector)", ex);
            }
            finally
            {
                if (document != null) Marshal.ReleaseComObject(document);
                if (item != null) Marshal.ReleaseComObject(item);
            }
            return summary;
        }
        private string GetSummaryFromExplorer(Outlook._Explorer explorer)
        {
            if (explorer == null) return null;
            FolderPath = GetFolderPath(explorer);
            string summary = GetSummaryFromInlineResponse(explorer);
            
            if (summary != null) return summary;
            Outlook.Selection selection = null;
            try
            {
                selection = explorer.Selection;
            }
            catch (System.Exception ex)
            {
                // skip this exception; it may fire when in certain folders such as RSS Feeds
            }
            if (selection != null)
            {
                summary = GetSummaryFromSelection(selection, explorer);
            }
            return summary;
        }

        private string GetSummaryFromSelection(Outlook.Selection selection, Outlook._Explorer explorer)
        {
            int hostVersion = MyOutlookAddin1.AddinModule.CurrentInstance.HostMajorVersion;

            string summary = null;
            object item = null;
            try
            {

                if (selection == null) return "No selection returned";
                if (selection.Count == 0) return "No selection";

                item = selection[1];
                if (selection.Count > 1)
                {
                    summary = "Multiple items selected";
                }
                else // a single item is selected
                {
                    summary = GetSummaryFromOutlookItem(item);
                }

                string selectedText = null;
                if (hostVersion > 15) //The PreviewPane object is only available in Outlook 2016+
                {
                    selectedText = GetTextSelectedInReadingPane(explorer);
                }
                else
                {
                    Outlook._Inspector inspector = null;
                    Word.Document document = null;
                    try
                    {
                        inspector =
                            item.GetType().InvokeMember("GetInspector", System.Reflection.BindingFlags.GetProperty, null, item, null) as Outlook._Inspector;
                        document = inspector.WordEditor as Word.Document;
                        selectedText = GetTextSelectedInDocument(document);
                    }
                    catch (System.Exception ex)
                    {
                        Logging.Log.Add("ContextInfo.GetSummaryFromExplorer. Trying to get text selected in the inspector of the first selected item; that is expected to be the Reading Pane", ex);
                    }
                    finally
                    {
                        if (inspector != null) Marshal.ReleaseComObject(inspector);
                        if (document != null) Marshal.ReleaseComObject(document);
                    }
                }

                if (selectedText == null)
                {
                    summary += "\nSelectedText=(none)";
                }
                else
                {
                    summary += "\nSelectedText='" + selectedText + "'";
                }
            }
            catch (System.Exception ex)
            {
                Logging.Log.Add("ContextInfo.GetSummaryFromExplorer(_Explorer)", ex);
            }
            finally
            {
                if (selection != null) Marshal.ReleaseComObject(selection);
                if (item != null) Marshal.ReleaseComObject(item);
            }
            return summary;
        }
        private string GetSummaryFromInlineResponse(Outlook._Explorer explorer)
        {
            string summary = null;

            if (MyOutlookAddin1.AddinModule.CurrentInstance.HostMajorVersion > 14) // if the add-in is loaded in Outlook 2013+
            {
                object inlineResponse = null;
                Word.Document document = null;
                try
                {
                    // use early binding if the add-in uses interops for Outlook 2013
                    // Outlook.MAPIFolder folder = explorer.ActiveInlineResponse; 

                    // use late binding since this add-in supports Outlook 2010+; see minOfficeVersionSupported in adxloader.dll.manifest and interops used
                    inlineResponse =
                        explorer.GetType().InvokeMember("ActiveInlineResponse", System.Reflection.BindingFlags.GetProperty, null, explorer, null);
                    if (inlineResponse != null)
                    {
                        summary = GetSummaryFromOutlookItem(inlineResponse);
                        WindowType = "Explorer|InlineResponse";
                        document =
                            explorer.GetType().InvokeMember("ActiveInlineResponseWordEditor", System.Reflection.BindingFlags.GetProperty, null, explorer, null) as Word.Document;
                        string selectedText = GetTextSelectedInDocument(document);                        
                        if (selectedText == null)
                        {
                            summary += "\nSelection: (none)";
                        }
                        else
                        {
                            summary += "\nSelection: " + selectedText;
                        }
                    }
                }
                catch (System.Exception ex)
                {
                    Logging.Log.Add("ContextInfo.GetInlineResponseInfo", ex);
                }
                finally
                {
                    if (document != null) Marshal.ReleaseComObject(document);
                    if (inlineResponse != null) Marshal.ReleaseComObject(inlineResponse);
                }
            }
            return summary;
        }
        private string GetSummaryFromOutlookItem(object item)
        {
            bool submitted = false;
            try
            {
                submitted = System.Convert.ToBoolean(item.GetType().InvokeMember("Submitted", System.Reflection.BindingFlags.GetProperty, null, item, null));
            }
            catch (System.Exception ex)
            {
                // skip. Assume the item doesn't have the Submitted property
            }
            string summary;
            if (submitted)
            {
                summary = "(item is submitted)";
            }
            else
            {
                // early binding
                // Outlook.MAPIFolder folder = mailItem.Subject; 

                // late binding
                string subject = item.GetType().InvokeMember("Subject", System.Reflection.BindingFlags.GetProperty, null, item, null).ToString();
                summary = "Subject='" + subject + "'" + "\n" + "Path='" + GetParentFolderPath(item) + "'";
            }
            return summary;
        }

        private string GetTextSelectedInReadingPane(Outlook._Explorer explorer)
        {
            string selectedText = null;

            if (MyOutlookAddin1.AddinModule.CurrentInstance.HostMajorVersion > 15) // if the add-in is loaded in Outlook 2016+
            {
                object previewPane = null;
                Word.Document document = null;
                try
                {
                    // use early binding if the add-in uses interops for Outlook 2016
                    // Outlook.MAPIFolder folder = explorer.PreviewPane; 

                    // use late binding since this add-in supports Outlook 2010+; see minOfficeVersionSupported in adxloader.dll.manifest and interops used
                    previewPane = explorer.GetType().InvokeMember("PreviewPane", System.Reflection.BindingFlags.GetProperty, null, explorer, null);
                    if (previewPane != null)
                    {
                        // early binding: 
                        // document = previewPane.WordEditor;

                        try
                        {
                            // late binding
                            document = previewPane.GetType().InvokeMember("WordEditor", System.Reflection.BindingFlags.GetProperty, null, previewPane, null) as Word.Document;
                        }
                        catch (System.Exception ex)
                        {
                            Logging.Log.Add("ContextInfo.GetReadingPaneSelectionInfo", "SKIP - previewPane.WordEditor exception");
                        }

                        selectedText = GetTextSelectedInDocument(document);
                        if (selectedText != null)
                        {
                            WindowType = "Explorer|ReadingPane";
                        }
                        Marshal.ReleaseComObject(previewPane);
                    }
                }
                catch (System.Exception ex)
                {
                    Logging.Log.Add("ContextInfo.GetReadingPaneSelectionInfo", ex);
                }
                finally
                {
                    if (previewPane != null) Marshal.ReleaseComObject(previewPane);
                    if (document != null) Marshal.ReleaseComObject(document);
                }
            }
            return selectedText;
        }
        private string GetTextSelectedInDocument(Word.Document document)
        {
            if (document == null) return null;
            Word.Windows windows = null;
            Word.Window window = null;
            Word.Selection selection = null;
            try
            {
                windows = document.Windows;
                if (windows.Count > 0)
                {
                    window = windows[1];
                    selection = window.Selection;
                    if (selection.Start == selection.End) return null;
                    return selection.Text;
                }
            }
            catch (System.Exception ex)
            {
                Logging.Log.Add("ContextInfo.GetWordSelection", ex);
            }
            finally
            {
                if (selection != null) Marshal.ReleaseComObject(selection);
                if (window != null) Marshal.ReleaseComObject(window);
                if (windows != null) Marshal.ReleaseComObject(windows);
            }
            return null;
        }

        private string GetFolderPath(Outlook._Explorer explorer)
        {
            Outlook.MAPIFolder folder = explorer.CurrentFolder;
            string folderPath = GetFolderPath(folder);
            Marshal.ReleaseComObject(folder);
            return folderPath;
        }

        public override string ToString()
        {
            return string.Format("Window: {0}\nCaption: {1}\n{2}", WindowType, WindowCaption, ContextDataString);
        }

        private string GetParentFolderPath(object item)
        {
            // early binding
            // Outlook.MAPIFolder folder = mailItem.Parent; 

            // late binding
            Outlook.MAPIFolder folder = item.GetType().InvokeMember("Parent", System.Reflection.BindingFlags.GetProperty, null, item, null) as Outlook.MAPIFolder;
            string folderPath = GetFolderPath(folder);
            Marshal.ReleaseComObject(folder);
            return folderPath;
        }

        public static string GetFolderPath(Outlook.MAPIFolder folder)
        {
            string folderPath = "";
            bool toBeReleased = false;
            object tempObj = null;
            while (folder != null)
            {
                folderPath = "\\" + folder.Name + folderPath;
                try
                {
                    tempObj = folder.Parent;
                }
                catch
                {
                    tempObj = null;
                }
                if (toBeReleased)
                {
                    if (folder != null) Marshal.ReleaseComObject(folder);
                }
                else // the caller will relese the folder passed
                    toBeReleased = true;
                //The parent of a root folder is of the Outlook.Namespace type
                folder = (tempObj as Outlook.MAPIFolder);
            }
            if (tempObj != null) Marshal.ReleaseComObject(tempObj);
            if (folderPath != "")
                folderPath = folderPath.Remove(0, 1);
            return folderPath;
        }
        public static string GetHostSummary()
        {
            MyOutlookAddin1.AddinModule addinModule = MyOutlookAddin1.AddinModule.CurrentInstance;
            string s = string.Format("HostBitness={0}\nHostMajorVersion={1}\nHostName={2}\n"
                + "HostStartMode={3}\nHostType={4}\nHostVersion={5}",
                addinModule.HostBitness, addinModule.HostMajorVersion, addinModule.HostName,
                addinModule.HostStartMode, addinModule.HostType, addinModule.HostVersion);
            return s;
        }

    }
}
